package ru.dbotthepony.mc.otm;

import net.minecraft.core.BlockPos;
import net.minecraft.world.level.Level;
import net.minecraftforge.event.TickEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.fmlserverevents.FMLServerStartedEvent;
import net.minecraftforge.fmlserverevents.FMLServerStoppingEvent;
import ru.dbotthepony.mc.otm.capability.IMatterHandler;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Supplier;

public class MatterGrid {
	public static final Set<MatterGrid> NETWORKS = new HashSet<>();

	private final Set<IMatterGridCell> cells = new HashSet<>();

	private static final Set<Supplier<Boolean>> discovering_neighbours = new HashSet<>();

	public static void scheduleDiscoverNeighbours(IMatterGridCell cell, BlockPos pos, Level level) {
		discovering_neighbours.add(() -> !cell.isValidMatterCell() || cell.connectOrCreateMatterGrid(pos, level, true));
	}

	@SubscribeEvent
	public void onServerStarting(FMLServerStartedEvent event) {
		discovering_neighbours.clear();
	}

	@SubscribeEvent
	public void onServerStopping(FMLServerStoppingEvent event) {
		discovering_neighbours.clear();
	}

	@SubscribeEvent
	public static void discoverNeighbours(TickEvent.ServerTickEvent event) {
		if (event.phase == TickEvent.Phase.END && discovering_neighbours.size() != 0) {
			ArrayList<Supplier<Boolean>> invalid = new ArrayList<>();

			for (Supplier<Boolean> cell : discovering_neighbours) {
				if (cell.get()) {
					invalid.add(cell);
				}
			}

			if (invalid.size() != 0) {
				for (Supplier<Boolean> cell : invalid) {
					discovering_neighbours.remove(cell);
				}
			}
		}
	}

	public MatterGrid() {
		NETWORKS.add(this);
	}

	public BigDecimal getCapacity() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				summ = summ.add(handler.getMaxStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal getPotentialCapacity() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				summ = summ.add(handler.getMaxStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal getStored() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				summ = summ.add(handler.getStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal getPotentialStored() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				summ = summ.add(handler.getStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal extractMatter(BigDecimal howMuch, boolean simulate) {
		if (howMuch.compareTo(BigDecimal.ZERO) <= 0)
			return BigDecimal.ZERO;

		BigDecimal extracted = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				BigDecimal drain = handler.extractMatterOuter(howMuch, simulate);

				if (!drain.equals(BigDecimal.ZERO)) {
					extracted = extracted.add(drain);
					howMuch = howMuch.subtract(drain);

					if (howMuch.compareTo(BigDecimal.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return extracted;
	}

	public BigDecimal softPushMatter(BigDecimal howMuch, boolean simulate) {
		if (howMuch.compareTo(BigDecimal.ZERO) <= 0)
			return BigDecimal.ZERO;

		BigDecimal received = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				BigDecimal receive = handler.receiveMatterOuter(howMuch, simulate);

				if (!receive.equals(BigDecimal.ZERO)) {
					received = received.add(receive);
					howMuch = howMuch.subtract(receive);

					if (howMuch.compareTo(BigDecimal.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return received;
	}

	public BigDecimal pushMatter(BigDecimal howMuch, boolean simulate) {
		if (howMuch.compareTo(BigDecimal.ZERO) <= 0)
			return BigDecimal.ZERO;

		BigDecimal received = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.EXTRACT) {
				BigDecimal receive = handler.receiveMatterOuter(howMuch, simulate);

				if (!receive.equals(BigDecimal.ZERO)) {
					received = received.add(receive);
					howMuch = howMuch.subtract(receive);

					if (howMuch.compareTo(BigDecimal.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return received;
	}

	public void track(IMatterGridCell entity) {
		if (cells.contains(entity))
			return;

		cells.add(entity);
		entity.setMatterGrid(this);

		// OverdriveThatMatters.LOGGER.debug("Tracking {} in {}. Tracking {} in total", entity, this, entities.size());
	}

	public void untrack(IMatterGridCell entity) {
		if (!cells.contains(entity))
			return;

		cells.remove(entity);
		entity.setMatterGrid(null);

		// OverdriveThatMatters.LOGGER.debug("Untracking {} in {}. Tracking {} in total", entity, this, entities.size());
	}

	public int networkSize() {
		return cells.size();
	}

	public void mergeWith(MatterGrid other) {
		if (other == this)
			return;

		MatterGrid a;
		MatterGrid b;

		if (networkSize() > other.networkSize()) {
			a = this;
			b = other;
		} else {
			a = other;
			b = this;
		}

		b.validate();

		for (IMatterGridCell entity : b.cells) {
			a.track(entity);
		}

		NETWORKS.remove(b);
	}

	public void validate() {
		ArrayList<IMatterGridCell> invalid = new ArrayList<>();

		for (IMatterGridCell entity : cells) {
			if (!entity.isValidMatterCell()) {
				invalid.add(entity);
			}
		}

		if (invalid.size() != 0) {
			for (IMatterGridCell entity : invalid) {
				untrack(entity);
			}
		}

		if (cells.size() == 0) {
			NETWORKS.remove(this);
		}
	}
}
